/*******************************************************************************
 *
 * MIT License
 *
 * Copyright 2025 AMD ROCm(TM) Software
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 *******************************************************************************/

#include <unordered_set>
#include <vector>

#include <rocRoller/KernelGraph/KernelGraph.hpp>

namespace rocRoller
{
    namespace KernelGraph
    {
        namespace RemoveSetCoordinateDetails
        {
            /**
             * Find the leave nodes (in terms of program order) inside the subgraph rooted
             * at a given node.
            */
            void findLeaves(int                      tag,
                            KernelGraph const&       kg,
                            std::unordered_set<int>& visited,
                            std::vector<int>&        leaves);

            /**
             * Find the leave nodes (in terms of program order) inside the subgraphs rooted
             * at a given set of nodes.
            */
            std::vector<int> findLeaves(std::vector<int> nodes, KernelGraph const& kg);

            /**
             * Remove all SetCoordinate nodes from a control graph of a given kernel graph
            */
            void removeSetCoordinates(KernelGraph& kg);
        }
    }
}
