/*******************************************************************************
 *
 * MIT License
 *
 * Copyright 2024-2025 AMD ROCm(TM) Software
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 *******************************************************************************/

/**
 * random number generator
 */

#pragma once

#include "Operation.hpp"

#include <memory>
#include <unordered_set>

namespace rocRoller
{
    namespace Operations
    {
        class RandomNumberGenerator : public BaseOperation
        {
        public:
            enum class SeedMode
            {
                Default, //< Use original value
                PerThread, //< Thread (workitem) ID will be added to seed
                Count
            };

            RandomNumberGenerator() = delete;
            explicit RandomNumberGenerator(OperationTag seed);
            RandomNumberGenerator(OperationTag seed, SeedMode mode);

            std::unordered_set<OperationTag> getInputs() const;
            std::string                      toString() const;
            SeedMode                         getSeedMode() const;

            OperationTag seed;
            SeedMode     mode;

            bool operator==(RandomNumberGenerator const&) const;
        };

        std::string   toString(RandomNumberGenerator::SeedMode const&);
        std::ostream& operator<<(std::ostream& stream, RandomNumberGenerator::SeedMode mode);
    }
}
